/*
 * Functions for elementary operations for
 * simplicial complex computations.
 *
 * by M.Hachimori, May 2000.
 */

#include <stdlib.h>
#include "s_complex.h"



/*
 * Function to enumerate all faces from
 * a list of facets.
 * The input list of facets are assumed to be sorted.
 * The output list is also sorted.
 */
Simplices *facets2faces(faces)
  Simplices *faces;
{
  Simplices *initial, *tail, *current, *output, *return_address, *tmp, *tmp2;
  int i, j, d, num_newface;

  if (faces == NULL)
    return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  return_address = output;
  output->prev = NULL;
  output->next = NULL;
  initial = output; /* 'initial' be the first and 'tail' be the last data of new d-dim faces.*/
  d = faces->vert[0];

  current = faces;    /* 'current' is a pointer sweeping the input data. */

  /* copy the largest dimension part */
  while(current != NULL && current->vert[0] == d){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output = output->next;
    for (i = 0; i <= current->vert[0]+1; i++)
      output->vert[i] = current->vert[i];
    output->next = NULL;
    current = current->next;
    }
  tail = output;
  d--;
  initial = initial->next; 

  while(output->vert[0] >= 1){
    num_newface = 0;
    /* copy 1-dim-lower faces to output.*/
    while (current!= NULL && current->vert[0] > d)
      current = current->next;
    while (current!= NULL && current->vert[0] == d){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      for (i = 0; i <= current->vert[0]+1; i++)
        output->next->vert[i] = current->vert[i];
      output->next->prev = output;
      output = output->next;
      output->next = NULL;
      num_newface ++;
      current = current->next;}

  
    /* compute codim 1 faces of faces between `initial' and `tail'
       and add them to output. */
    while(initial != NULL && initial->vert[0] == d+1){
      for (i = 1; i <= initial->vert[0]+1; i ++){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        output->vert[0] = initial->vert[0] - 1;
        for (j = 1; j <= initial->vert[0]+1; j++) /* list all codim 1 faces */
          if (j < i) output->vert[j] = initial->vert[j];
          else if (j > i) output->vert[j-1] = initial->vert[j];
        num_newface ++;}
      initial = initial->next;
      }

    /* sort the faces and remove multiplicity before going to the next step*/
    initial = tail->next;
    msort(initial, num_newface);
    tmp = initial;
    while(tmp->next != NULL){
      if (cmp_faces(tmp->vert, (tmp->next)->vert) == 0){
        tmp2 = (tmp->next)->next;
        free(tmp->next);
        tmp->next = tmp2;
        if(tmp->next != NULL)
          tmp->next->prev = tmp;
        }
      else
        tmp = tmp->next;}
    tail = tmp;
    output = tmp;

    d --;
    }

  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;
  return return_address;
}


/*
 * Function to calculate a set of facets
 * from a set of faces.
 */
Simplices *faces2facets(faces)
  Simplices *faces;
{
  Simplices *return_address, *head, *tail, *current, *output, *tmp;
  int i, j, d, mark;

  if (faces == NULL) return NULL;


  d = faces->vert[0];
  output = (Simplices*)malloc(sizeof(Simplices));
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  current = faces;
  head = current;
  /* All the largest dimensional faces are copied to output. */
  while(current!= NULL && current->vert[0] == d){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output = output->next;
    output->next = NULL;
    for (i = 0; i <= current->vert[0]+1; i ++)
      output->vert[i] = current->vert[i];
    current = current->next;
    }
  d --;

  /* For the smaller dimensional faces, check whether they are
     contained in a larger face or not. */
  while(d >= 0){
    while(current != NULL && current->vert[0] == d) {
      tmp = head;
      mark = 0;
      while(tmp != NULL && tmp->vert[0] > d) {
        if (face_contained(current->vert, tmp->vert) == 1){
          mark = 1;
          break;}
        tmp = tmp->next;
        }
      if (mark == 0){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        for (i = 0; i <= current->vert[0]+1; i ++)
          output->vert[i] = current->vert[i];
        }
      current = current->next;
      }

    d --;
  }

  /* remove the first extra data before returning. */
  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;
  return return_address;
}



/*
 * Function to calculate boundary complex
 *   the input is a list of facets and
 *   the output is also a list of facets.
 */
Simplices *boundary_facets(facets)
  Simplices *facets;
{
  Simplices *return_address, *output, *head;
  Simplices *tmp, *tmp1, *tmp2;
  Simplices *top_added, *current1, *current2;
  int i, j, d, num_newface;

  if (facets == NULL || facets->vert[0] <= 0)
     return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  output->next = NULL;
  return_address = output;
          /* This first data will not be used.
             Just defined here to adjust for the looping
             and removed in the last of this function. */
  d = facets->vert[0];

  while(d > 0){
    head = output;
    num_newface = 0;
    /* first list up all codim 1 faces of d-dim facets. */
    while(facets != NULL && facets->vert[0] == d){
      for (i = 1; i <= facets->vert[0]+1; i ++){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        output->vert[0] = facets->vert[0]-1;
        for (j = 1; j <= facets->vert[0]+1; j ++)
          if (j < i) output->vert[j] = facets->vert[j];
          else if (j > i) output->vert[j-1] = facets->vert[j];
        num_newface ++;
        }
      facets = facets->next;}
    msort(head->next, num_newface);

    /* if there are more than one identical faces, then
       remove all of them.*/
    top_added = head->next;
    while(head->next != NULL){
      if (cmp_faces(head->vert, head->next->vert) == 0)
        if (head->next->next != NULL &&
            cmp_faces(head->next->vert, head->next->next->vert) == 0){
          tmp = head->next->next;
          free(head->next);
          head->next = tmp;
          head->next->prev = head;}
        else
          if (head->prev != NULL){
            tmp1 = head->prev;
            tmp2 = head->next->next;
            free(head->next);
            free(head);
            if (tmp1 != NULL) tmp1->next = tmp2;
            if (tmp2 != NULL) tmp2->prev = tmp1;
            head = tmp2;
            if (head == NULL) break;}
          else{
            tmp2 = head->next->next;
            free(head->next);
            free(head);
            tmp2->prev = NULL;
            head = tmp2;}
      else
        head = head->next;
      }

    /* if d is not the largest dimension, then remove
       faces which is contained in other larger face. */
    current1 = top_added; /* top_added is the top face of added faces.*/
    while(current1 != NULL){
      current2 = return_address->next; /* current2 sweeps all the larger
                                         faces in the output. */
      if (current2 == NULL) break; /* if no boundary, break. */
      while(current2->vert[0] >= d){
        if (face_contained(current1->vert, current2->vert) == 1){
          tmp = current1->prev;
          tmp->next = current1->next;
          current1->next->prev = tmp;
          free(current1);
          current1 = tmp->next;
          continue;}
        current2 = current2->next;}
      current1 = current1->next;
      }

    d --;
  }

  /* remove the null head */
  if (return_address->next != NULL){
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}
  else{
    free(return_address);
    return NULL;}

}


/*
 * Function to calculate k-th pure part.
 *   Input = set of faces
 *   Output = set of facets of k-th pure part
 */
Simplices *pure_part(faces, k)
  Simplices *faces;
  int k;
{
  Simplices *output, *return_address;
  int i;

  if (faces == NULL || faces->vert[0] < k)
    return NULL;

  output = (Simplices*)malloc(sizeof(Simplices)); /* this first data will not
                                                     be used and will be removed
                                                     in the last of this function.*/
  output->prev = NULL;
  return_address = output;

  /* search the first face whose dimension is k */
  while(faces->vert[0] > k)
    faces = faces->next;

  /* output all k-dim faces */
  while(faces != NULL && faces->vert[0] == k){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output->next->next = NULL;
    output = output->next;
    for (i = 0; i <= faces->vert[0]+1; i ++)
      output->vert[i] = faces->vert[i];
    faces = faces->next;
    }

  /* remove the first extra data */
  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;

  return return_address;

}


/*
 * Function to calculate f-vector.
 *   Input is a set of faces
 *   Output is a vector whose i-th element is f_(i-1)
 *   (Note that the index is shifted by 1.
 *    If the input is NULL, then return f_vec[0] = 1,
 *    just show that there is only one (-1)-dim face.)
 */
void get_fvector(faces, f_vec)
  Simplices *faces;
  int f_vec[];
{
  int i;

  if (faces == NULL){
    f_vec[0] = 1;
    return;}

  f_vec[0] = 1;
  /* initially set f_vec[i] = 0 for all i>=1. */
  for (i = 1; i <= faces->vert[0]+1; i ++)
    f_vec[i] = 0;

  /* count the number of faces. */
  while(faces != NULL){
    f_vec[faces->vert[0]+1] ++;
    faces = faces->next;
    }

  return;
}


/*
 * Function to calculate h-vector from f-vector.
 */
void f_to_h(fvec, hvec, dim)
  int fvec[];
  int hvec[];
  int dim;
{
  int i, j;
  int tmp[MAXDIM+2], tmp2[MAXDIM+2];

  tmp[0] = fvec[0];
  for (i = 0; i <= dim+1; i ++)
    {
    tmp2[0] = 1;
    for (j = 1; j <= i; j ++)
      tmp2[j] = tmp[j] - tmp[j - 1];
    tmp2[i + 1] = fvec[i + 1];
    for (j = 0; j <= i + 1; j ++)
      tmp[j] = tmp2[j];
    }
  for(i = 0; i <= dim+1; i ++)
    hvec[i] = tmp[i];
  return;
}


/*
 * Function to check whether a vector is an M-sequence or not.
 *   dim = dimension = (number of entries -1)
 */
int Mseq(hvec, dim)
  int hvec[];
  int dim;
{
  short i;
  static int M(int n, int i);

    /*
     * function used in Mseq.
     */
    static int M(n, i)
      int n;
      int i;
    {
      int j, k;
      int ans, bi[MAXDIM+1];
    
      binom_expand(n, i, bi);
      k = 1;
      while(bi[k] == 0)
        k ++;
      ans = 0;
      for (j = k; j <= i; j ++)
        ans += binom(bi[j] + 1, j + 1);
    
      return ans;
    }

  if (hvec[0] != 1)
    return 0;
  for (i = 1; i <= dim; i ++)
    if (hvec[i] < 0 || (hvec[i] > 0 && M(hvec[i], i) < hvec[i + 1]))
      return 0;
  if (hvec[dim+1] < 0)
    return 0;
  return 1;
}

/*
 * function to calculate binomial coefficient
 */
int binom(n, r)
  int n;
  int r;
{
  int i;
  double ans;

  ans = 1;
  for (i = n - r + 1; i <= n; i ++)
    ans *= i;
  for (i = 2; i <= r; i ++)
    ans /= i;

  return (int)ans;
}


/*
 * function to calculate binomial expansion
 */
void binom_expand(n, i, bi)
  int n;
  int i;
  int bi[];
{
  int j;
  int tmp;

  while(n > 0)
    {
    tmp = i;
    while (binom(tmp + 1, i) <= n)
      tmp ++;
    bi[i] = tmp;
    n -= binom(tmp, i);
    i --;
    }
  for (j = 1; j <= i; j ++)
    bi[j] = j-1;
  bi[0] = 0;

  return;
}


/*
 * Function to calculate intersection of two simplicial complexes.
 *   Input: two lists of faces.
 *   Output: a list of faces of their intersection
 */
Simplices *intersection_faces(faces1, faces2)
  Simplices *faces1;
  Simplices *faces2;
{
  Simplices *output, *return_address, *current, *faces2_head;
  int i, tmp;

  if (faces1 == NULL || faces2 == NULL) return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
                                /* this first data will be removed
                                   in the last of this function. */
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  /* for each face in faces1, check whether it is in faces2 and
     copy to output if in. */
  while(faces1 != NULL){
    /* first adjust the pointer of second complex
       to the dimension of the face "faces1". */
    while (faces2 != NULL & faces2->vert[0] > faces1->vert[0])
      faces2 = faces2->next;
    if (faces2 == NULL) break;

    current = faces2;
    while(current != NULL && current->vert[0] == faces1->vert[0]){
      if (cmp_faces(faces1->vert, current->vert) == 0){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output->next->next = NULL;
        output = output->next;
        for (i = 0; i <= faces1->vert[0]+1; i ++)
          output->vert[i] = faces1->vert[i];
        break;}
      current = current->next;}
  faces1 = faces1->next;}

  /* remove the first extra data before returning the answer. */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}

}
Simplices *intersection_faces_no2(faces1, faces2)
  Simplices *faces1;
  Simplices *faces2;
{
  Simplices *output, *return_address, *current, *faces2_head;
  int i, tmp;

  if (faces1 == NULL || faces2 == NULL) return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
                                /* this first data will be removed
                                   in the last of this function. */
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  /* for each face in faces1, check whether it is in faces2 and
     copy to output if in. */
  while(faces1 != NULL){
      if (search(faces1->vert, faces2) == 1){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output->next->next = NULL;
        output = output->next;
        for (i = 0; i <= faces1->vert[0]+1; i ++)
          output->vert[i] = faces1->vert[i];}
  faces1 = faces1->next;}

  /* remove the first extra data before returning the answer. */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}

}

/*
 * Function to calculate the union of two simplicial complexes.
 *   Input: two lists of faces
 *   Output: list of faces of their union.
 */
Simplices *union_faces(faces1, faces2)
  Simplices *faces1;
  Simplices *faces2;
{
  Simplices *output, *return_address, *current, *tail, *head_output;
  int i, tmp, num_newface;

  if (faces1 == NULL && faces2 == NULL)
    return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  /* first, copy all faces from faces2 whose dimension is larger
     than faces1. */
  while(faces1 == NULL ||
        (faces2 != NULL && faces2->vert[0] > faces1->vert[0])){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output->next->next = NULL;
    output = output->next;
    for (i = 0; i <= faces2->vert[0]+1; i ++)
      output->vert[i] = faces2->vert[i];
    faces2 = faces2->next;}

  while(faces1 != NULL){
    num_newface = 0;
    head_output = output;
    /* copy faces of faces1 of dim k. */
    current = faces1;
    while(current != NULL && current->vert[0] == faces1->vert[0]){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      output->next->prev = output;
      output->next->next = NULL;
      output = output->next;
      for (i = 0; i <= current->vert[0]+1; i ++)
        output->vert[i] = current->vert[i];
      num_newface ++;
      current = current->next;}
    tail = current;

    /* copy faces of faces2 of dim k which is not already output */
    while(faces2 != NULL && faces2->vert[0] == faces1->vert[0]){
      tmp = 0;
      current = faces1;
      while(current != NULL && current != tail){
        if (cmp_faces(faces2->vert, current->vert) == 0){
          tmp = 1; break;}
        current = current->next;}
      if (tmp == 0){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output->next->next = NULL;
        output = output->next;
        for (i = 0; i <= faces1->vert[0]+1; i ++)
          output->vert[i] = faces2->vert[i];
        num_newface ++;
        }
      faces2 = faces2->next;}

    faces1 = tail;
    if (num_newface > 0)
      msort(head_output->next, num_newface);
    }

  /* remove the first extra data */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}


}


/*
 * Function to calculate the (open) star of a face in a simplicial complex.
 *   Input: list of faces and one face
 *   Output: list of faces of the star of the face
 */
Simplices *star_faces(faces, face)
  Simplices *faces;
  short face[];
{
  Simplices *output, *return_address;
  int i;

  if (faces == NULL) return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
                                    /* remove this extra data
                                       in the end of this function. */
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  while(faces != NULL){
    if (face[0] == -1 || face_contained(face, faces->vert) == 1){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      output->next->prev = output;
      output->next->next = NULL;
      output = output->next;
      for (i = 0; i <= faces->vert[0]+1; i ++)
        output->vert[i] = faces->vert[i];
      }
    faces = faces->next;}

  /* remove the first extra data */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}


}

/*
 * Function to calculate the link of a face in a simplicial complex.
 *   Input: list of faces and one face
 *   Output: list of faces of the link of the face
 * (The algorithm is the same as calculating open star
 *  except for the part of the output part.)
 */
Simplices *link_faces(faces, face)
  Simplices *faces;
  short face[];
{
  Simplices *output, *return_address;
  int i, j, k, tmp;

  if (faces == NULL || faces->vert[0] <= face[0]) return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  while(faces != NULL){
    if (face[0] == -1 ||
        (face[0] < faces->vert[0] && face_contained(face, faces->vert) == 1)){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      output->next->prev = output;
      output->next->next = NULL;
      output = output->next;
      /* output is "faces->vert" minus "face". */
      output->vert[0] = faces->vert[0] - face[0] - 1;
      k = 1;
      for (i = 1; i <= faces->vert[0]+1; i ++){
        tmp = 0;
        for (j = 1; j <= face[0]+1; j ++)
          if (faces->vert[i] == face[j]){
            tmp = 1;
            break;}
        if (tmp == 0)
          output->vert[k ++] = faces->vert[i];
        }
      }
    faces = faces->next;}

  /* remove the first extra data */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}

}




/*
 * Function to calculate the deletion of a face in a simplicial complex.
 *   Input: list of faces and one face
 *   Output: list of faces of the deletion of the face
 */
Simplices *deletion_faces(faces, face)
  Simplices *faces;
  short face[];
{
  Simplices *output, *return_address;
  int i;

  if (faces == NULL) return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  while(faces != NULL){
    if (face_contained(face, faces->vert) == 0){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      output->next->prev = output;
      output->next->next = NULL;
      output = output->next;
      for (i = 0; i <= faces->vert[0]+1; i ++)
        output->vert[i] = faces->vert[i];
      }
    faces = faces->next;}


  /* remove the first extra data */
  if (return_address->next == NULL){
    free(return_address);
    return NULL;}
  else{
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}
}


/*
 * Function to calculate stellar subdivision
 *   Input: list of faces and a face
 *   Output: list of faces of stellar subdivision.
 * (Assuming that "face" is a face of "faces".
 *  So this fact should be checked before using this function.)
 */
Simplices *stellar_subdivision(faces, face)
  Simplices *faces;
  short face[];
{
  Simplices *output, *return_address, *vertices, *head, *current;
  Simplices *star, *clstar, *lk, *lk_head, *tmp;
  int i, newvertex, num_face, dim;

  if (faces == NULL) return NULL;
  if (face[0] == 0)    /* if "face" is 0-dim, just return a copy of "faces". */
    return copy_faces(faces);

  output = deletion_faces(faces, face);
  star = star_faces(faces, face);
  clstar = facets2faces(star); delete_faces(star);
  lk = deletion_faces(clstar, face); delete_faces(clstar);
  lk_head = lk;
    /* "lk" is (closed star - open star) = (link * [boundary of face]) */

  /* determine the index of new vertex introduced */
  vertices = pure_part(faces, 0);
  newvertex = 0;
  current = vertices;
  while(current != NULL){
    if (current->vert[1] >= newvertex)
      newvertex = current->vert[1] + 1;
    current = current->next;}
  delete_faces(vertices);

  /* add newly introduced faces (lk * newvertex) */
  output->prev = (Simplices*)malloc(sizeof(Simplices)); /* extra first data added */
  output->prev->next = output;
  output = output->prev;
  output->prev = NULL;
  return_address = output;
  while(output->next != NULL && output->next->vert[0] > lk->vert[0]+1)
    output = output->next;    /* Now "output" is set to the last face
                                 of the larger dimension than "lk". */
  dim = lk->vert[0];
  while(dim >= 0){
    head = output;
    num_face = 0;
    /* move output to the end of (dim+1)-faces */
    while(output->next != NULL && output->next->vert[0] >= dim + 1){
      output = output->next;
      num_face ++;}
    /* add new faces */
    while(lk != NULL && lk->vert[0] == dim){
      tmp = output->next;
      output->next = (Simplices*)malloc(sizeof(Simplices));
      output->next->prev = output;
      output->next->next = tmp;
      output = output->next;
      output->next->prev = output;
      output->vert[0] = dim + 1;
      for (i = 1; i <= dim+1; i ++)
        output->vert[i] = lk->vert[i];
      output->vert[dim+2] = newvertex;
      num_face ++;
      lk = lk->next;}
    msort(head->next, num_face);  /* sort the dim-faces */
    dim --;}
  delete_faces(lk_head);
  /* Lastly, add the newly introduced vertex to the output. */
  while(output->next != NULL)
    output = output->next;
  output->next = (Simplices*)malloc(sizeof(Simplices));
  output->next->prev = output;
  output = output->next;
  output->next = NULL;
  output->vert[0] = 0; output->vert[1] = newvertex;

  /* remove the extra data in the first of the list*/
  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;

  return return_address;

}


/*
 * Function to caltulate barycentric subdivision
 *   Input: a list of faces
 *   Output: a list of faces of barycentric subdivision
 */
Simplices *barycentric_subdivision(faces)
  Simplices *faces;
{
  Simplices *current, *subdivision, *tmp;

  subdivision = faces;
  current = faces;
  while(current != NULL && current->vert[0] >= 1){
    tmp = stellar_subdivision(subdivision, current->vert);
    if (subdivision != faces) delete_faces(subdivision);
    subdivision = tmp;
    current = current->next;}

  return subdivision;
}


/*
 * Function to calculate join of two complexes
 *   Input: two lists of faces
 *   Output: the list of faces of the join of two complexes.
 *
 * This function assumes that two complexes have no common vertices.
 * So checking before calling this function is needed.
 */
Simplices *join_faces(faces1, faces2)
  Simplices *faces1;
  Simplices *faces2;
{

  Simplices *output[MAXDIM+1];
  Simplices *current1, *current2, *head[MAXDIM+1];
  int num_newface[MAXDIM+1];
  int i, idx1, idx2, d, dim;


  if (faces1 == NULL && faces2 == NULL) return NULL;

  if (faces1 == NULL)    /* if "faces1" is NULL, just return a copy of "faces2". */
    return copy_faces(faces2);

  if (faces2 == NULL)    /* if "faces2" is NULL, just return a copy of "faces1". */
    return copy_faces(faces1);


  /* if both are not NULL, calculate join. */
  /* initial setting */
  dim = faces1->vert[0] + faces2->vert[0] + 1;
  for (i = 0; i <= dim; i ++)
    num_newface[i] = 0;
  for (i = 0; i <= dim; i ++){
    output[i] = (Simplices*)malloc(sizeof(Simplices));
    output[i]->prev = NULL;
    output[i]->next = NULL;
    head[i] = output[i];}

  /* compute joined faces classified by dimension d. */
  /* First, compute the joins of two nonempty faces. */
  current1 = faces1;
  while(current1 != NULL){
    current2 = faces2;
    while(current2 != NULL){
      d = current1->vert[0] + current2->vert[0] + 1;
      output[d]->next = (Simplices*)malloc(sizeof(Simplices));
      output[d]->next->prev = output[d];
      output[d] = output[d]->next;
      output[d]->next = NULL;
      output[d]->vert[0] = d;
      idx1 = 1; idx2 = 1;
      for (i = 1; i <= d+1; i ++)
        if (idx1 <= current1->vert[0]+1 &&
            current1->vert[idx1] < current2->vert[idx2])
          output[d]->vert[i] = current1->vert[idx1 ++];
        else
          output[d]->vert[i] = current2->vert[idx2 ++];
      current2 = current2->next;
      num_newface[d] ++;}
    current1 = current1->next;}

   /* Next, the cases that one is empty face. */
   for (i = 0; i <= faces1->vert[0]; i ++){
     output[i]->next = pure_part(faces1, i);
     while(output[i]->next != NULL){
     output[i] = output[i]->next;
     num_newface[i] ++;}
     }
   for (i = 0; i <= faces2->vert[0]; i ++){
     output[i]->next = pure_part(faces2, i);
     while(output[i]->next != NULL){
     output[i] = output[i]->next;
     num_newface[i] ++;}
     }

   /* sort joined faces */
  for (i = 0; i <= dim; i ++)
    msort(head[i]->next, num_newface[i]);

  /* merge the faces into one list
     (Remark: each dimension always has at least one newly added face
              because we are assuming that two input lists are lists
              of faces.
              If lists of facets are allowed to be input, this merging
              part should be improved.) */
  head[dim] = head[dim]->next;
  free(head[dim]->prev);
  head[dim]->prev = NULL;
  for (i = dim; i >= 1; i --){
    head[i-1] = head[i-1]->next;
    free(head[i-1]->prev);
    output[i]->next = head[i-1];
    head[i-1]->prev = output[i];}

  return head[dim];

}


/*
 * Function to do bistellar exchange.
 * Assume that the input is a pseudomanifold.
 *   Input: list of faces and a face.
 *   Output: list of faces of bistellar exchange.
 *
Simplices *bistellar_exchange(faces, face) */

/*
 * Function to check purity.
 *
 *   Input: list of faces.
 *   Return 1 if pure, else return 0.
 */
int purity_check(faces)
  Simplices *faces;
{
  Simplices *facets, *current;

  if (faces == NULL) return 1;

  facets = faces2facets(faces);

  current = facets;
  while(current->next != NULL){
    if (current->vert[0] != current->next->vert[0]){
      free(facets);
      return 0;}
    current = current->next;}
  
  /* free(facets); */
  delete_faces(facets);
  return 1;
}

/*
 * Function to check PMness
 *
 *   Input: list of faces.
 *   Return 1 if PM, else return 0.
 *
 * Remark: Assuming that input is pure.
 *         If the data is not known to be pure, use purity_check().
 */
int pm_check(faces)
  Simplices *faces;
{
  Simplices *ridges, *current, *ridges_head;
  int dim, num;

  dim = faces->vert[0];
  ridges = pure_part(faces, dim-1);
  ridges_head = ridges;

  while(ridges != NULL){
    current = faces;
    num = 0;
    while(current != NULL && current->vert[0] == dim){
      if (face_contained(ridges->vert, current->vert) == 1)
        if (num ++ >= 2){
          free(ridges_head);
          return 0;}
      current = current->next;}
    ridges = ridges->next;}
  free(ridges_head);
  return 1;

}


/*
 * Function to operate bistellar exchange.
 *   Input: a list of faces of a pseudomanifold and a face
 *   Modify the list into a new list after bistellar exchange.
 *
 * Assume that the input is a pseudomanifold and the face
 * admits a bistellar exchange.
 * Checking before using this function is needed.
 */
Simplices *bistellar_exchange(faces, a)
  Simplices *faces;
  short a[MAXDIM+2];
{
  Simplices *b, *b_faces, *a_facet, *a_boundary, *lk, *lk_vert;
  Simplices *current, *vertices;
  Simplices  *deletion, *join, *unionn;
  int i, newvertex;

  if (a[0] < faces->vert[0]){ /* if a is not a facet, no new vertex needed. */
    /* calculate b which is a nonface s.t. lk(a) = bd(b). */
    b = (Simplices*)malloc(sizeof(Simplices));
    b->prev = NULL; b->next = NULL;
    lk = link_faces(faces, a);
    lk_vert = pure_part(lk, 0);
    b->vert[0] = -1;
    current = lk_vert;
    while(current != NULL){
      b->vert[++(b->vert[0]) + 1] = current->vert[1];
      current = current->next;}
    b_faces = facets2faces(b);
    free(b);
  
    /* calculate the boundary of a. */
    a_facet = (Simplices*)malloc(sizeof(Simplices));
    a_facet->prev = NULL; a_facet->next = NULL;
    for (i = 0; i <= a[0]+1; i ++)
      a_facet->vert[i] = a[i];
    a_boundary = facets2faces(a_facet);
    a_boundary = a_boundary->next;
    free(a_boundary->prev);
    a_boundary->prev = NULL;

    /* delete a from complex */
    deletion = deletion_faces(faces, a);

    /* add bd(a)*b */
    join = join_faces(a_boundary, b_faces);
    delete_faces(a_boundary);
    delete_faces(b_faces);
    unionn = union_faces(deletion, join);
    delete_faces(deletion);
    delete_faces(join);

    return unionn;}

  else{ /* if a is a facet, return stellar subdivision by a. */
    return stellar_subdivision(faces, a);}

}

Simplices *intersection_facets(facets1, facets2)
  Simplices *facets1;
  Simplices *facets2;
{
  Simplices *faces1, *faces2, *output;

  faces1 = facets2faces(facets1);
  faces2 = facets2faces(facets2);

  output = intersection_faces(faces1, faces2);

  delete_faces(faces1);
  delete_faces(faces2);
  return (output);

}

int M_check_facets(facets)
  Simplices *facets;
{
  Simplices *faces;
  int fvec[MAXDIM+1], hvec[MAXDIM+1];
  int dim;

  faces = facets2faces(facets);
  dim = faces->vert[0];
  get_fvector(faces, fvec);
  (void)delete_faces(faces);
  f_to_h(fvec, hvec, dim);
  return (Mseq(hvec, dim));
}


/* Function to list edges which is contractible.
   : if the vertices in the link of an edge vw and
     the vertices jointed to both v and w are the same,
     the edge vw is contractible. */
int contractible_edges(faces)
  Simplices *faces;
{
  Simplices *link;
  Simplices *faces_head, *faces_start;
  Simplices *edges, *edges_head;
  short face1[MAXDIM+2], face2[MAXDIM+2];
  int dim, contract, tmp, i, j;

  if (faces == NULL) return;
  faces_head = faces;
  dim = faces->vert[0];
  if (dim <= 0) return;
  edges = pure_part(faces, 1); edges_head = edges;

/*  while(faces->vert[0] == dim) */
  while(faces->vert[0] > 0)
    faces = faces->next;
  faces_start = faces;

  while(edges != NULL){
    link = link_faces(faces_head, edges->vert);
    contract = 0;
    faces = faces_start;
    while(faces != NULL){
      if (search(faces->vert, link)) {
        faces = faces->next; continue;}
      /* face1 is v*face, face2 is w*face. */
      face1[0] = faces->vert[0] + 1;
      tmp = 0;
      for (i = 1; i <= face1[0]; i++){
        if (faces->vert[i] < edges->vert[1])
          face1[i] = faces->vert[i];
        else if (faces->vert[i] == edges->vert[1]){
          tmp = 1; break;}
        else if (faces->vert[i] > edges->vert[1]){
          break;}
        }
      if (tmp == 1) {faces = faces->next; continue;}
      face1[i] = edges->vert[1];
      for (j = i+1; j <= face1[0]+1; j++)
        face1[j] = faces->vert[j-1];

      face2[0] = faces->vert[0] + 1;
      tmp = 0;
      for (i = 1; i <= face2[0]; i++){
        if (faces->vert[i] < edges->vert[2])
          face2[i] = faces->vert[i];
        else if (faces->vert[i] == edges->vert[2]){
          tmp = 1; break;}
        else if (faces->vert[i] > edges->vert[2]){
          break;}
        }
      if (tmp == 1) {faces = faces->next; continue;}
      face2[i] = edges->vert[2];
      for (j = i+1; j <= face2[0]+1; j++)
        face2[j] = faces->vert[j-1];

/*
printf("edge=%d %d: ", edges->vert[1], edges->vert[2]);
printf("face=(%d) %d %d %d %d: ", faces->vert[0], faces->vert[1], faces->vert[2], faces->vert[3], faces->vert[4]);
printf("face1=(%d) %d %d %d %d: ", face1[0], face1[1], face1[2], face1[3], face1[4]);
printf("face2=(%d) %d %d %d %d\n", face2[0], face2[1], face2[2], face2[3], face2[4]);
*/

      if (search(face1, faces_head) && search(face2, faces_head)){
        contract = 1; break;}
      faces = faces->next;
      }
    if (contract == 0)
      printf("%d %d is contractible\n", edges->vert[1], edges->vert[2]);
else
      printf("%d %d is not contractible\n", edges->vert[1], edges->vert[2]);

    delete_faces(link);
    edges = edges->next;
    }

  delete_faces(edges_head);

  return;
}

